import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
import h5py
from typing import Dict

import logging

logger = logging.getLogger(__name__)

from pegasusio import modalities, UnimodalData, CITESeqData, MultimodalData


def load_10x_h5_file_v2(h5_in: h5py.Group) -> MultimodalData:
    """Load 10x v2 format matrix from hdf5 file

    Parameters
    ----------

    h5_in : h5py.Group
        An instance of h5py.Group class that is connected to a 10x v2 formatted hdf5 file.

    Returns
    -------

    A MultimodalData object containing (genome, UnimodalData) pair per genome.

    Examples
    --------
    >>> io.load_10x_h5_file_v2(h5_in)
    """
    data = MultimodalData()
    for genome in h5_in.keys():
        group = h5_in[genome]

        M, N = group["shape"][...]
        mat = csr_matrix(
            (
                group["data"][...],
                group["indices"][...],
                group["indptr"][...],
            ),
            shape=(N, M),
        )

        barcodes = group["barcodes"][...].astype(str)
        ids = group["genes"][...].astype(str)
        names = group["gene_names"][...].astype(str)

        unidata = UnimodalData(
            {"barcodekey": barcodes},
            {"featurekey": names, "featureid": ids},
            {"counts": mat},
            {"modality": "rna", "genome": genome},
            cur_matrix = "counts",
        )
        unidata.separate_channels()

        data.add_data(unidata)

    return data


def load_10x_h5_file_v3(h5_in: h5py.Group) -> MultimodalData:
    """Load 10x v3 format matrix from hdf5 file, allowing detection of crispr and citeseq libraries

    Parameters
    ----------

    h5_in : h5py.Group
        An instance of h5py.Group class that is connected to a 10x v3 formatted hdf5 file.

    Returns
    -------

    A MultimodalData object containing (genome, UnimodalData) pair per genome.

    Examples
    --------
    >>> io.load_10x_h5_file_v3(h5_in)
    """
    M, N = h5_in["matrix/shape"][...]
    bigmat = csr_matrix(
        (
            h5_in["matrix/data"][...],
            h5_in["matrix/indices"][...],
            h5_in["matrix/indptr"][...],
        ),
        shape=(N, M),
    )
    barcodes = h5_in["matrix/barcodes"][...].astype(str)
    df = pd.DataFrame(
        data={
            "genome": h5_in["matrix/features/genome"][...].astype(str),
            "feature_type": h5_in["matrix/features/feature_type"][...].astype(str),
            "id": h5_in["matrix/features/id"][...].astype(str),
            "name": h5_in["matrix/features/name"][...].astype(str),
        }
    )

    genomes = list(df["genome"].unique())
    if "" in genomes:
        genomes.remove("")
    default_genome = genomes[0] if len(genomes) == 1 else None

    data = MultimodalData()
    gb = df.groupby(by=["genome", "feature_type"])
    for name, group in gb:
        barcode_metadata = {"barcodekey": barcodes}
        feature_metadata = {
            "featurekey": group["name"].values,
            "featureid": group["id"].values,
        }
        mat = bigmat[:, gb.groups[name]]

        genome = (
            name[0] if (name[0] != "" or default_genome is None) else default_genome
        )
        modality = "custom"
        if name[1] == "Gene Expression":
            modality = "rna"
        elif name[1] == "CRISPR Guide Capture":
            modality = "crispr"
        elif name[1] == "Antibody Capture":
            modality = "citeseq"

        Class = CITESeqData if modality == "citeseq" else UnimodalData
        unidata = Class(
            barcode_metadata,
            feature_metadata,
            {"counts": mat},
            {"genome": genome, "modality": modality},
            cur_matrix = "counts",
        )
        unidata.separate_channels()

        data.add_data(unidata)

    return data


def load_10x_h5_file(input_h5: str) -> MultimodalData:
    """Load 10x format matrix (either v2 or v3) from hdf5 file

    Parameters
    ----------

    input_h5 : `str`
        The matrix in 10x v2 or v3 hdf5 format.

    Returns
    -------

    A MultimodalData object containing (genome, UnimodalData) pair per genome.

    Examples
    --------
    >>> io.load_10x_h5_file('example_10x.h5')
    """
    data = None
    with h5py.File(input_h5, "r") as h5_in:
        load_file = (
            load_10x_h5_file_v3 if "matrix" in h5_in.keys() else load_10x_h5_file_v2
        )
        data = load_file(h5_in)

    return data


def load_loom_file(
    input_loom: str, genome: str = None, modality: str = None
) -> MultimodalData:
    """Load count matrix from a LOOM file.

    Parameters
    ----------

    input_loom : `str`
        The LOOM file, containing the count matrix.
    genome : `str`, optional (default None)
        The genome reference. If None, use "unknown" instead. If not None and input loom contains genome attribute, the attribute will be overwritten.
    modality: `str`, optional (default None)
        Modality. If None, use "rna" instead. If not None and input loom contains modality attribute, the attribute will be overwritten.

    Returns
    -------

    A MultimodalData object containing a (genome, UmimodalData) pair.

    Examples
    --------
    >>> io.load_loom_file('example.loom', genome = 'GRCh38')
    """
    col_trans = {
        "CellID": "barcodekey",
        "obs_names": "barcodekey",
        "cell_names": "barcodekey",
    }
    row_trans = {
        "Gene": "featurekey",
        "var_names": "featurekey",
        "gene_names": "featurekey",
        "Accession": "featureid",
        "gene_ids": "featureid",
        "ensembl_ids": "featureid",
    }

    import loompy

    with loompy.connect(input_loom) as ds:
        barcode_metadata = {}
        barcode_multiarrays = {}
        for key, arr in ds.col_attrs.items():
            key = col_trans.get(key, key)
            if arr.ndim == 1:
                barcode_metadata[key] = arr
            elif arr.ndim > 1:
                barcode_multiarrays[key] = arr
            else:
                raise ValueError(
                    f"Detected column attribute '{key}' has ndim = {arr.ndim}!"
                )

        feature_metadata = {}
        feature_multiarrays = {}
        for key, arr in ds.row_attrs.items():
            key = row_trans.get(key, key)
            if arr.ndim == 1:
                feature_metadata[key] = arr
            elif arr.ndim > 1:
                feature_multiarrays[key] = arr
            else:
                raise ValueError(
                    f"Detected row attribute '{key}' has ndim = {arr.ndim}!"
                )

        barcode_multigraphs = {}
        for key, graph in ds.col_graphs.items():
            barcode_multigraphs[key] = csr_matrix(graph)
        feature_multigraphs = {}
        for key, graph in ds.row_graphs.items():
            feature_multigraphs[key] = csr_matrix(graph)

        matrices = {}
        for key, mat in ds.layers.items():
            key = "X" if key == "" else key
            matrices[key] = mat.sparse().T.tocsr()

        metadata = dict(ds.attrs)
        if genome is not None:
            metadata["genome"] = genome
        elif "genome" not in metadata:
            metadata["genome"] = "unknown"

        if modality is not None:
            metadata["modality"] = modality
        elif "modality" not in metadata:
            if metadata.get("experiment_type", "none") in modalities:
                metadata["modality"] = metadata.pop("experiment_type")
            else:
                metadata["modality"] = "rna"

        unidata = UnimodalData(
            barcode_metadata,
            feature_metadata,
            matrices,
            metadata,
            barcode_multiarrays,
            feature_multiarrays,
            barcode_multigraphs,
            feature_multigraphs,
            cur_matrix = "X",
        )
        unidata.separate_channels()

    data = MultimodalData(unidata)
    return data


def write_loom_file(data: MultimodalData, output_file: str) -> None:
    """Write a MultimodalData to loom file. Will assert data only contain one type of experiment. Use current matrix as the main matrix."""
    keys = data.list_data()
    if len(keys) > 1:
        raise ValueError(f"Data contain multiple modalities: {','.join(keys)}!")
    data.select_data(keys[0])
    matrices = data.list_keys()
    main_mat_key = data.current_matrix()
    if len(matrices) == 0:
        raise ValueError("Could not write empty matrix to a loom file!")

    def _replace_slash(name: str) -> str:
        """Replace slash with |"""
        if name.find("/") >= 0:
            return name.replace("/", "|")
        return name

    def _process_attrs(
        key_name: str, attrs: pd.DataFrame, attrs_multi: dict
    ) -> Dict[str, object]:
        res_dict = {key_name: attrs.index.values}
        for key in attrs.columns:
            res_dict[_replace_slash(key)] = np.array(attrs[key].values)
        for key, value in attrs_multi.items():
            if (
                value.ndim > 1
            ):  # value.ndim == 1 refers to np.recarray, which will not be written to a loom file.
                res_dict[_replace_slash(key)] = (
                    value if value.shape[1] > 1 else value[:, 0]
                )
        return res_dict

    row_attrs = _process_attrs("Gene", data.var, data.varm)
    col_attrs = _process_attrs("CellID", data.obs, data.obsm)

    accession_key = (
        "featureid"
        if "featureid" in row_attrs
        else ("gene_ids" if "gene_ids" in row_attrs else None)
    )
    if accession_key is not None:
        row_attrs["Accession"] = row_attrs.pop(accession_key)

    layers = {}
    for matkey in matrices:
        layers["" if matkey == main_mat_key else matkey] = data.get_matrix(matkey).T

    file_attrs = {}
    for key, value in data.uns.items():
        if isinstance(value, str):
            file_attrs[_replace_slash(key)] = value

    import loompy

    loompy.create(output_file, layers, row_attrs, col_attrs, file_attrs=file_attrs)

    if len(data.varp) > 0 or len(data.obsp) > 0:
        with loompy.connect(output_file) as ds:
            for key, value in data.varp.items():
                ds.row_graphs[_replace_slash(key)] = value
            for key, value in data.obsp.items():
                ds.col_graphs[_replace_slash(key)] = value

    logger.info(f"{output_file} is written.")


def write_10x_h5(data: MultimodalData, output_file: str) -> None:
    """Follow 10x hdf5 format description (https://support.10xgenomics.com/single-cell-gene-expression/software/pipelines/latest/advanced/h5_matrices).
    Restricted to GEX modality only.
    Settings consistent with 10x cellranger's (https://github.com/10XGenomics/cellranger/blob/master/lib/python/cellranger/matrix.py#L428-L445).
    """
    unidata = data._unidata
    nmat = len(unidata.matrices)
    compression_method = "gzip"
    chunk_size = 80000

    assert unidata.get_modality() == "rna", "Only Gene Expression count matrix is accepted!"

    def _create_h5_string_dataset(group, name, data, shape=None, fillvalue=None):
        """Inspired by 10x cellranger create_hdf5_string_dataset function (https://github.com/10XGenomics/cellranger/blob/master/lib/python/cellranger/io.py#L323-L358)"""
        kwargs = {
            'chunks': (chunk_size,),
            'maxshape': (None,),
            'compression': compression_method,
            'shuffle': True,
        }

        if data is None:
            assert fillvalue is not None, "Either data or fillvalue must be specified!"
            dtype = f"S{len(fillvalue)}"
            group.create_dataset(
                name=name,
                dtype=dtype,
                shape=shape,
                fillvalue=fillvalue.encode('ascii', 'xmlcharrefreplace'),
                **kwargs,
            )
        else:
            fixed_len = max(len(x) for x in data)
            if fixed_len == 0:
                fixed_len = 1
            dtype = f"S{fixed_len}"
            group.create_dataset(
                name=name,
                data=np.vectorize(lambda s: s.encode('ascii', 'xmlcharrefreplace'))(data),
                dtype=dtype,
                **kwargs,
            )

    def _write_h5(data, output_file):
        n_obs, n_feature = data.shape
        genome = data.uns["genome"] if "genome" in data.uns.keys() else "Unknown"
        feature_type = "Gene Expression"
        with h5py.File(output_file, "w") as f:
            grp = f.create_group("matrix")

            # Cell barcodes
            _create_h5_string_dataset(grp, "barcodes", data.obs_names.values)

            # Raw counts.
            X = data.X if isinstance(data.X, csr_matrix) else csr_matrix(data.X)
            grp.create_dataset(
                name="data",
                data=X.data,
                chunks=(chunk_size,),
                maxshape=(None,),
                compression=compression_method,
                shuffle=True,
            )
            grp.create_dataset(
                name="indices",
                data=X.indices,
                chunks=(chunk_size,),
                maxshape=(None,),
                compression=compression_method,
                shuffle=True,
            )
            grp.create_dataset(
                name="indptr",
                data=X.indptr,
                chunks=(chunk_size,),
                maxshape=(None,),
                compression=compression_method,
                shuffle=True,
            )
            grp.create_dataset(
                name="shape",
                dtype=np.int32,
                data=(n_feature, n_obs),
            )  # feature-by-barcode

            feature_grp = grp.create_group("features")
            feature_grp.create_dataset(name="_all_tag_keys", data=[b"genome"])
            _create_h5_string_dataset(feature_grp, "feature_type", data=None, shape=(n_feature,), fillvalue=feature_type)
            _create_h5_string_dataset(feature_grp, "genome", data=None, shape=(n_feature,), fillvalue=genome)
            _create_h5_string_dataset(feature_grp, "id", data=data.var["featureid"].values)
            _create_h5_string_dataset(feature_grp, "name", data=data.var_names.values)

    if nmat == 1:
        _write_h5(unidata, output_file)
    else:
        import os

        path = os.path.dirname(os.path.abspath(output_file))
        outname = os.path.basename(output_file).rstrip(".h5")
        for mat_key in unidata.list_keys():
            unidata.select_matrix(mat_key)
            _write_h5(unidata, f"{path}/{outname}.{mat_key}.h5")



import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from typing import Dict, Optional, Union

import logging

from pegasusio.unimodal_data import UnimodalData

logger = logging.getLogger(__name__)


class SpatialData(UnimodalData):
    """
    Class to implement data structure to
    manipulate spatial data with the spatial image (img) field
    This class extends UnimodalData with additional
    functions specific to the img field
    """

    def __init__(
        self,
        barcode_metadata: Optional[Union[dict, pd.DataFrame]] = None,
        feature_metadata: Optional[Union[dict, pd.DataFrame]] = None,
        matrices: Optional[Dict[str, csr_matrix]] = None,
        metadata: Optional[dict] = None,
        barcode_multiarrays: Optional[Dict[str, np.ndarray]] = None,
        feature_multiarrays: Optional[Dict[str, np.ndarray]] = None,
        barcode_multigraphs: Optional[Dict[str, csr_matrix]] = None,
        feature_multigraphs: Optional[Dict[str, csr_matrix]] = None,
        cur_matrix: str = "X",
        image_metadata: Optional[pd.DataFrame] = None,
    ) -> None:
        assert metadata["modality"] == "visium"
        super().__init__(
            barcode_metadata,
            feature_metadata,
            matrices,
            metadata,
            barcode_multiarrays,
            feature_multiarrays,
            barcode_multigraphs,
            feature_multigraphs,
            cur_matrix,
        )
        self.image_metadata = image_metadata

    @property
    def img(self) -> Optional[pd.DataFrame]:
        return self.image_metadata

    @img.setter
    def img(self, img: pd.DataFrame):
        self.image_metadata = img

    def __repr__(self) -> str:
        repr_str = super().__repr__()
        key = "img"
        fstr = self._gen_repr_str_for_attrs(key)
        if fstr != "":
            repr_str += f"\n    {key}: {fstr}"
        return repr_str



import os, re

import pandas as pd

from pegasusio import MultimodalData, SpatialData
from .hdf5_utils import load_10x_h5_file
import json
from PIL import Image


def process_spatial_metadata(df):
    df["in_tissue"] = df["in_tissue"].apply(lambda n: True if n == 1 else False)
    df["barcodekey"] = df["barcodekey"].map(lambda s: s.split("-")[0])
    df.set_index("barcodekey", inplace=True)


def load_visium_folder(input_path) -> MultimodalData:
    """
    Method to read the visium spatial data folder
    into MultimodalData object that contains SpatialData
    """
    file_list = os.listdir(input_path)
    sample_id = input_path.split("/")[-1]
    # Load count matrix.
    hdf5_filename = "filtered_feature_bc_matrix.h5"
    assert hdf5_filename in file_list, "Raw count hdf5 file is missing!"
    rna_data = load_10x_h5_file(f"{input_path}/{hdf5_filename}")

    # Load spatial metadata.
    assert ("spatial" in file_list) and (
        os.path.isdir(f"{input_path}/spatial")
    ), "Spatial folder is missing!"
    tissue_pos_csv = "spatial/tissue_positions_list.csv"
    spatial_metadata = pd.read_csv(
        f"{input_path}/{tissue_pos_csv}",
        names=[
            "barcodekey",
            "in_tissue",
            "array_row",
            "array_col",
            "pxl_col_in_fullres",
            "pxl_row_in_fullres",
        ],
    )
    process_spatial_metadata(spatial_metadata)

    barcode_metadata = rna_data.obs.join(spatial_metadata, how='left')
    feature_metadata = rna_data.var

    matrices = {"X": rna_data.X}
    metadata = {"genome": rna_data.get_genome(), "modality": "visium"}

    #  Store “pxl_col_in_fullres” and ”pxl_row_in_fullres” as a 2D array,
    # which is the spatial location info of each cell in the dataset.
    spatial_coords = barcode_metadata[['pxl_row_in_fullres', 'pxl_col_in_fullres']]
    barcode_multiarrays = {"X_spatial": spatial_coords.to_numpy()}
    barcode_metadata.drop(columns=['pxl_row_in_fullres', 'pxl_col_in_fullres'], inplace=True)

    # Store image metadata as a Pandas DataFrame, with the following structure:
    image_metadata = pd.DataFrame()
    spatial_path = f"{input_path}/spatial"

    with open(f"{spatial_path}/scalefactors_json.json") as fp:
        scale_factors = json.load(fp)

    def get_image_data(filepath, sample_id, image_id, scaleFactor, spot_diameter_fullres):
        data = Image.open(filepath)
        dict = {
            "sample_id": sample_id,
            "image_id": image_id,
            "data": data,
            "scale_factor": scaleFactor,
            "spot_diameter": spot_diameter_fullres * scaleFactor,
        }
        return dict

    for png in [f for f in os.listdir(spatial_path) if re.match(".*\.png", f)]:
        if ("_hires_" in png) or ("_lowres_" in png):
            filepath = f"{spatial_path}/{png}"
            res_tag = "hires" if "_hires_" in png else "lowres"
            image_item = get_image_data(
                filepath,
                sample_id,
                res_tag,
                scale_factors[f"tissue_{res_tag}_scalef"],
                scale_factors["spot_diameter_fullres"]
            )
            image_metadata = image_metadata.append(image_item, ignore_index=True)

    assert not image_metadata.empty, "the image data frame is empty"
    spdata = SpatialData(
        barcode_metadata,
        feature_metadata,
        matrices,
        metadata,
        barcode_multiarrays=barcode_multiarrays,
        image_metadata=image_metadata,
    )
    data = MultimodalData(spdata)

    return data



# This code is inspired by HCA zarr python codes and AnnData zarr codes

import os
import shutil
import PIL
import numpy as np
import pandas as pd
from pandas.api.types import is_categorical_dtype, is_string_dtype, is_scalar, is_dict_like
from scipy.sparse import csr_matrix, issparse
import zarr
from zarr import Blosc
from natsort import natsorted
from typing import List, Dict, Tuple, Union

import logging
logger = logging.getLogger(__name__)

from pegasusio import modalities
from pegasusio import UnimodalData, VDJData, CITESeqData, CytoData, MultimodalData, SpatialData


CHUNKSIZE = 1000000.0
COMPRESSOR = Blosc(cname = 'lz4', clevel = 5)


def calc_chunk(shape: tuple) -> tuple:
    ndim = len(shape)
    chunks = [1] * ndim
    ords = np.argsort(shape)
    if shape[ords[0]] > 0:
        chunk_size = CHUNKSIZE
        chunk_value = -1
        for i, idx in enumerate(ords):
            if chunk_value < 0:
                if shape[idx] ** (ndim - i) < chunk_size:
                    chunks[idx] = shape[idx]
                    chunk_size /= chunks[idx]
                else:
                    chunk_value = int(np.ceil(chunk_size ** (1.0 / (ndim - i))))
                    chunks[idx] = chunk_value
            else:
                chunks[idx] = chunk_value
    return tuple(chunks)


def check_and_remove_existing_path(path: str) -> None:
    if os.path.exists(path):
        if os.path.isdir(path):
            shutil.rmtree(path)
            logger.warning(f"Detected and removed pre-existing directory {path}.")
        else:
            assert os.path.isfile(path)
            os.unlink(path)
            logger.warning(f"Detected and removed pre-existing file {path}.")


class ZarrFile:
    def __init__(self, path: str, mode: str = 'r', storage_type: str = None) -> None:
        """ Initialize a Zarr file, if mode == 'w', create an empty one, otherwise, load from path
        path : `str`, path for the zarr object.
        storage_type : `str`, currently only support 'ZipStore' and 'NestedDirectoryStore'. If None, use 'NestedDirectoryStore' by default.
        """
        self.store = self.root = None
        self.write_empty_chunks = False

        if storage_type is None:
            storage_type = 'NestedDirectoryStore'

        if mode == 'w':
            # Create a new zarr file
            check_and_remove_existing_path(path)
            self.store = zarr.ZipStore(path, mode = 'w') if storage_type == 'ZipStore' else zarr.NestedDirectoryStore(path)
            self.root = zarr.group(self.store, overwrite = True)
            self.write_empty_chunks = (storage_type == 'ZipStore')
        else:
            # Load existing zarr file
            self.store = zarr.NestedDirectoryStore(path) if os.path.isdir(path) else zarr.ZipStore(path, mode = 'r')
            if mode == 'a' and isinstance(self.store, zarr.ZipStore):
                self._to_directory()
            self.root = zarr.open(self.store, mode = mode)


    def __del__(self):
        if self.store is not None and hasattr(self.store, 'close'):
            self.store.close()


    def _to_zip(self):
        if not isinstance(self.store, zarr.ZipStore):
            zip_path = self.store.path + '.zip'
            zip_store = zarr.ZipStore(zip_path, mode = 'w')
            zarr.copy_store(self.store, zip_store)
            zip_store.close()

            shutil.rmtree(self.store.path)

            self.store = zarr.ZipStore(zip_path, mode = 'r')
            self.root = zarr.open_group(self.store, mode = 'r')


    def _to_directory(self):
        orig_path = self.store.path

        if not orig_path.endswith('.zip'):
            self.store.close()
            zip_path = orig_path + '.zip'
            check_and_remove_existing_path(zip_path)
            os.replace(orig_path, zip_path)
            self.store = zarr.ZipStore(zip_path, mode = 'r')
        else:
            zip_path = orig_path

        dest_path = zip_path[:-4]
        check_and_remove_existing_path(dest_path)
        dir_store = zarr.NestedDirectoryStore(dest_path)
        zarr.copy_store(self.store, dir_store)
        self.store.close()
        os.remove(zip_path)

        self.store = dir_store
        self.root = zarr.open_group(self.store)

        logger.info(f"Converted ZipStore zarr file {orig_path} to NestedDirectoryStore {dest_path}.")

    def read_csr(self, group: zarr.Group) -> csr_matrix:
        return csr_matrix((group['data'][...], group['indices'][...], group['indptr'][...]), shape = group.attrs['shape'])

    def read_series(self, group: zarr.Group, name: str) -> Union[pd.Categorical, np.ndarray]:
        if 'ordered' in group[name].attrs:
            # categorical column
            return pd.Categorical.from_codes(group[name][...], categories = group[f'_categories/{name}'][...], ordered = group[name].attrs['ordered'])
        else:
            if isinstance(group[name], zarr.hierarchy.Group):
                ll = []
                for data in group[name].arrays():
                    ll.append(PIL.Image.fromarray(data[1][...]))
                return ll
            else:
                return group[name][...]

    def read_dataframe(self, group: zarr.Group) -> pd.DataFrame:
        columns = group.attrs.get('columns', None)
        if columns is None:
            columns = [col for col in group.array_keys() if col != '_index']
        data = {col: self.read_series(group, col) for col in columns}
        _index = self.read_series(group, '_index')
        index = pd.Index(_index, name = group.attrs['index_name'], dtype = _index.dtype)
        df = pd.DataFrame(data = data, index = index) # if add columns = columns, the generation will be slow
        return df

    def read_array(self, group: zarr.Group, name: str) -> np.ndarray:
        return group[name][...]

    def read_record_array(self, group: zarr.Group) -> np.recarray:
        columns = group.attrs.get('columns', None)
        if columns is None:
            columns = [col for col in group.array_keys()]

        array = np.rec.fromarrays([group[col][...] for col in columns], names = columns)
        return array


    def read_composite_list(self, group: zarr.Group) -> list:
        assert '_size' in group.attrs
        res_list = [None] * group.attrs['_size']

        if 'scalar' in group.attrs:
            scalar_dict = group.attrs['scalar']
            for i, value in scalar_dict.items():
                res_list[int(i)] = value

        for key in group. array_keys():
            res_list[int(key)] = self.read_array(group, key)

        for key in group.group_keys():
            sub_group = group[key]
            data_type = sub_group.attrs['data_type']
            value = None
            if data_type == 'data_frame':
                value = self.read_dataframe(sub_group)
            elif data_type == 'record_array':
                value = self.read_record_array(sub_group)
            elif data_type == 'csr_matrix':
                value = self.read_csr(sub_group)
            elif data_type == 'dict':
                value = self.read_mapping(sub_group)
            else:
                assert data_type == 'composite_list'
                value = self.read_composite_list(sub_group)
            res_list[int(key)] = value

        return res_list


    def read_mapping(self, group: zarr.Group) -> dict:
        res_dict = {}

        if 'scalar' in group.attrs:
            res_dict.update(group.attrs['scalar'])

        for key in group.array_keys():
            res_dict[key] = self.read_array(group, key)

        for key in group.group_keys():
            sub_group = group[key]
            data_type = sub_group.attrs['data_type']
            value = None
            if data_type == 'data_frame':
                value = self.read_dataframe(sub_group)
            elif data_type == 'record_array':
                value = self.read_record_array(sub_group)
            elif data_type == 'csr_matrix':
                value = self.read_csr(sub_group)
            elif data_type == 'dict':
                value = self.read_mapping(sub_group)
            else:
                assert data_type == 'composite_list'
                value = self.read_composite_list(sub_group)
            res_dict[key] = value

        return res_dict


    def read_unimodal_data(self, group: zarr.Group) -> UnimodalData:
        """ Read UnimodalData
        """
        metadata = self.read_mapping(group['metadata'])
        assert 'genome' in metadata
        if 'modality' not in metadata:
            assert metadata.get('experiment_type', 'none') in modalities
            metadata['modality'] = metadata.pop('experiment_type')

        DataClass = UnimodalData
        modality = metadata['modality']
        if modality == 'tcr' or modality == 'bcr':
            DataClass = VDJData
        elif modality == 'citeseq':
            DataClass = CITESeqData
        elif modality == 'cyto':
            DataClass = CytoData
        elif modality == 'visium':
            DataClass = SpatialData

        unidata = DataClass(
            barcode_metadata=self.read_dataframe(group["barcode_metadata"]),
            feature_metadata=self.read_dataframe(group["feature_metadata"]),
            matrices=self.read_mapping(group["matrices"]),
            metadata=metadata,
            barcode_multiarrays=self.read_mapping(group["barcode_multiarrays"]),
            feature_multiarrays=self.read_mapping(group["feature_multiarrays"]),
            barcode_multigraphs=self.read_mapping(group["barcode_multigraphs"])
            if "barcode_multigraphs" in group
            else dict(),  # for backward-compatibility
            feature_multigraphs=self.read_mapping(group["feature_multigraphs"])
            if "feature_multigraphs" in group
            else dict(),
        )
        if isinstance (unidata, SpatialData):
            unidata.image_metadata = self.read_dataframe(group["image_metadata"]) if "image_metadata" in group else dict()

        if group.attrs.get("_cur_matrix", None) is not None:
            unidata.select_matrix(group.attrs["_cur_matrix"])

        return unidata


    def read_multimodal_data(self, attach_zarrobj = False) -> MultimodalData:
        """ Read MultimodalData
        """
        data = MultimodalData()
        for key, group in self.root.groups():
            unidata = self.read_unimodal_data(group)
            data.add_data(unidata)

        if self.root.attrs.get('_selected', None) is not None:
            data.select_data(self.root.attrs['_selected'])

        if attach_zarrobj:
            data._zarrobj = self

        return data


    def write_csr(self, parent: zarr.Group, name: str, matrix: csr_matrix) -> None:
        group = parent.create_group(name, overwrite = True)
        group.attrs.update(data_type = 'csr_matrix', shape = matrix.shape)
        group.create_dataset('data', data = matrix.data, shape = matrix.data.shape, chunks = calc_chunk(matrix.data.shape), dtype = matrix.data.dtype, compressor = COMPRESSOR, overwrite = True, write_empty_chunks = self.write_empty_chunks)
        group.create_dataset('indices', data = matrix.indices, shape = matrix.indices.shape, chunks = calc_chunk(matrix.indices.shape), dtype = matrix.indices.dtype, compressor = COMPRESSOR, overwrite = True, write_empty_chunks = self.write_empty_chunks)
        group.create_dataset('indptr', data = matrix.indptr, shape = matrix.indptr.shape, chunks = calc_chunk(matrix.indptr.shape), dtype = matrix.indptr.dtype, compressor = COMPRESSOR, overwrite = True, write_empty_chunks = self.write_empty_chunks)


    def write_series(self, group: zarr.Group, name: str, array: np.ndarray) -> None:
        if not is_categorical_dtype(array) and name != '_index' and is_string_dtype(array):
            keywords = set(array)
            if len(keywords) <= array.size / 10.0: # at least 10x reduction
                array = pd.Categorical(array, categories = natsorted(keywords))

        if is_categorical_dtype(array):
            # write category keys
            categories = group.require_group('_categories', overwrite = False)
            values = array.categories.values
            if isinstance(values[0], bytes):
                values = np.array([x.decode() for x in values], dtype = object)
            self.write_array(categories, name, values)
            # write codes
            codes_arr = group.create_dataset(name, data = array.codes, shape = array.codes.shape, chunks = calc_chunk(array.codes.shape), dtype = array.codes.dtype, compressor = COMPRESSOR, overwrite = True, write_empty_chunks = self.write_empty_chunks)
            codes_arr.attrs['ordered'] = array.ordered
        else:
            self.write_array(group, name, array)


    def write_dataframe(self, parent: zarr.Group, name: str, df: pd.DataFrame) -> None:
        group = parent.create_group(name, overwrite = True)
        attrs_dict = {'data_type' : 'data_frame', 'columns' : list(df.columns)}
        attrs_dict['index_name'] = df.index.name if df.index.name is not None else 'index'
        self.write_series(group, '_index', df.index.values)
        for col in df.columns:
            if df[col].size > 0 and isinstance(df[col].values[0], PIL.Image.Image):
                colgroup = group.create_group(col, overwrite = True)
                x = 0
                for data in df[col].values:
                    x = x+1
                    self.write_series(colgroup, col + str(x), np.array(data))
            else:
                self.write_series(group, col, df[col].values)
        group.attrs.update(**attrs_dict)

    def write_array(self, group: zarr.Group, name: str, array: np.ndarray) -> None:
        dtype = str if array.dtype.kind == 'O' else array.dtype
        group.create_dataset(name, data = array, shape = array.shape, chunks = calc_chunk(array.shape), dtype = dtype, compressor = COMPRESSOR, overwrite = True, write_empty_chunks = self.write_empty_chunks)

    def write_record_array(self, parent: zarr.Group, name: str, array: np.recarray) -> None:
        group = parent.create_group(name, overwrite = True)
        attrs_dict = {'data_type' : 'record_array', 'columns' : list(array.dtype.names)}
        for col in array.dtype.names:
            self.write_array(group, col, array[col])
        group.attrs.update(**attrs_dict)


    def write_composite_list(self, parent: zarr.Group, name: str, values: Union[list, tuple]) -> None:
        group = parent.create_group(name, overwrite = True)
        scalar_dict = {}

        for i, value in enumerate(values):
            if is_scalar(value):
                if type(value).__module__ == 'numpy':
                    value = value.item()
                scalar_dict[str(i)] = value
            elif isinstance(value, np.ndarray):
                if value.dtype.kind == 'V':
                    self.write_record_array(group, str(i), value)
                else:
                    self.write_array(group, str(i), value)
            elif isinstance(value, pd.DataFrame):
                self.write_dataframe(group, str(i), value)
            elif is_dict_like(value):
                self.write_mapping(group, str(i), value)
            elif issparse(value):
                assert isinstance(value, csr_matrix)
                self.write_csr(group, str(i), value)
            else:
                # assume value is either list or tuple
                assert type(value) in {list, tuple}
                is_comp_list = sum([is_scalar(x) for x in value]) < len(value)
                if is_comp_list:
                    self.write_composite_list(group, str(i), value)
                else:
                    # converting it to np.ndarray
                    self.write_array(group, str(i), value.astype(str) if is_categorical_dtype(value) else np.array(value))

        attrs_dict = {'data_type': 'composite_list', '_size': len(values)}
        if len(scalar_dict) > 0:
            attrs_dict['scalar'] = scalar_dict
        group.attrs.update(**attrs_dict)


    def write_mapping(self, parent: zarr.Group, name: str, mapping: dict, overwrite = True) -> None:
        group = None
        if overwrite:
            group = parent.create_group(name, overwrite = True)
        else:
            group = parent.require_group(name, overwrite = False) # throw an error if name in array_keys()

        scalar_dict = group.attrs.pop('scalar', {})

        def _write_one_pair(key, value):
            if is_scalar(value):
                if type(value).__module__ == 'numpy':
                    value = value.item()
                scalar_dict[key] = value
            elif isinstance(value, np.ndarray):
                if value.dtype.kind == 'V':
                    self.write_record_array(group, key, value)
                else:
                    self.write_array(group, key, value)
            elif isinstance(value, pd.DataFrame):
                self.write_dataframe(group, key, value)
            elif is_dict_like(value):
                self.write_mapping(group, key, value)
            elif issparse(value):
                assert isinstance(value, csr_matrix)
                self.write_csr(group, key, value)
            else:
                # assume value is either list or tuple
                assert type(value) in {list, tuple}
                is_comp_list = sum([is_scalar(x) for x in value]) < len(value)
                if is_comp_list:
                    self.write_composite_list(group, key, value)
                else:
                    # converting it to np.ndarray
                    self.write_array(group, key, value.astype(str) if is_categorical_dtype(value) else np.array(value))

        if overwrite:
            for key, value in mapping.items():
                _write_one_pair(key, value)
        else:
            for key in mapping.deleted:
                if key in scalar_dict:
                    del scalar_dict[key]
                else:
                    del group[key]
            for key in mapping.modified:
                _write_one_pair(key, mapping[key])

        attrs_dict = {'data_type' : 'dict'}
        if len(scalar_dict) > 0:
            attrs_dict['scalar'] = scalar_dict
        group.attrs.update(**attrs_dict)


    def write_unimodal_data(self, parent: zarr.Group, name: str, data: UnimodalData, overwrite: bool = True) -> None:
        """ Write UnimodalData
            overwrite means if overwrite the whole unimodal data; meaning is different from require_group overwrite --- if the name in array_keys(), take it and make it as a group
        """
        group = None
        if overwrite:
            group = parent.create_group(name, overwrite = True)
        else:
            group = parent.require_group(name, overwrite = False) # throw an error if name in array_keys()

        attrs_dict = {'data_type': 'UnimodalData', '_cur_matrix': data.current_matrix()}
        group.attrs.update(**attrs_dict)

        self.write_dataframe(group, 'barcode_metadata', data.barcode_metadata)
        self.write_dataframe(group, 'feature_metadata', data.feature_metadata)

        if hasattr(data, 'img'):
            self.write_dataframe(group, 'image_metadata', data.image_metadata)

        if overwrite or data.matrices.is_dirty():
            self.write_mapping(group, 'matrices', data.matrices, overwrite = overwrite)
        if overwrite or data.metadata.is_dirty():
            self.write_mapping(group, 'metadata', data.metadata, overwrite = overwrite)
        if overwrite or data.barcode_multiarrays.is_dirty():
            self.write_mapping(group, 'barcode_multiarrays', data.barcode_multiarrays, overwrite = overwrite)
        if overwrite or data.feature_multiarrays.is_dirty():
            self.write_mapping(group, 'feature_multiarrays', data.feature_multiarrays, overwrite = overwrite)
        if overwrite or data.barcode_multigraphs.is_dirty():
            self.write_mapping(group, 'barcode_multigraphs', data.barcode_multigraphs, overwrite = overwrite)
        if overwrite or data.feature_multigraphs.is_dirty():
            self.write_mapping(group, 'feature_multigraphs', data.feature_multigraphs, overwrite = overwrite)


    def write_multimodal_data(self, data: MultimodalData, overwrite: bool = True) -> None:
        """ Write MultimodalData
        """
        if overwrite:
            for key in data.list_data():
                self.write_unimodal_data(self.root, key, data.get_data(key), overwrite = True)
        else:
            for key in data.data.deleted:
                del self.root[key]
            for key in data.data.accessed:
                self.write_unimodal_data(self.root, key, data.get_data(key), overwrite = key in data.data.modified)
        self.root.attrs['_selected'] = data._selected
